from django.apps import AppConfig
from django.test import TestCase
from django.test.utils import override_settings

import affiliations
from affiliations.apps import AffiliateTrackingConfig
from affiliations.models import Partner


class TestAffiliateTrackingConfig(TestCase):

    @override_settings(
        AFFILIATE_TRIGGERS=[
            (
                'Partner updated',
                'django.db.models.signals.post_save',
                'affiliations.tests.utils.partner_updated',
                'affiliations.models.Partner',
            )
        ]
    )
    def setUp(self):
        self.affiliate_tracking_config = AffiliateTrackingConfig(
            'affiliations',
            affiliations
        )
        self.affiliate_tracking_config.ready()

    def test_name(self):
        self.assertEqual(
            self.affiliate_tracking_config.name,
            'affiliations'
        )
        self.assertTrue(issubclass(AffiliateTrackingConfig, AppConfig))

    def test_ready(self):
        partner = Partner.objects.create()
        partner.name = 'Partner updated'
        partner.save()

        partner.refresh_from_db()
        # The active = True is set by the receiver that was created for this
        # test
        self.assertTrue(partner.active)



"""Admin configuration for the application affiliations."""

from django.contrib import admin

from .models import (
    Partner,
    Subscription,
    Visitor,
)

admin.site.register(Partner)
admin.site.register(Subscription)
admin.site.register(Visitor)



"""Tests for the affiliations application."""
from .test_models import *
from .test_apps import *
from .test_affiliations_settings import *
from .test_middleware import *
from .test_utils import *
from .test_triggers import *



"""Configuration of the application affiliations."""

from __future__ import unicode_literals

from functools import partial

from django.apps import AppConfig
from django.conf import settings

from .utils import import_from_path


class AffiliateTrackingConfig(AppConfig):
    """Class representing the affiliations Django application."""

    name = 'affiliations'

    def ready(self):
        """Connect the signals from AFFILIATE_TRIGGERS."""
        for (
                name,
                signal_path,
                receiver_path,
                sender_path
        ) in settings.AFFILIATE_TRIGGERS:
            signal = import_from_path(signal_path)
            receiver = import_from_path(receiver_path)
            sender = import_from_path(sender_path)
            signal.connect(
                partial(
                    receiver,
                    affiliate_trigger_name=name
                ),
                sender=sender,
                weak=False
            )



"""This utils module will contain the utilities needed for the affiliations."""

from __future__ import absolute_import
try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse
try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode
import importlib

from .affiliations_settings import (
    AFFILIATE_QUERY_STRING_KEY,
)


def strip_affiliate_part_from_url(url):
    """
    Take care of stripping away the affiliate part of an URL.

    returns the stripped URL. Remove the ? if AFFILIATE_QUERY_STRING_KEY is the
    only query string parameter.

    An example of what is meant by "strip away":
    The request has the URL:
        https://example.com/some/page?page_id=9&affiliate_id=1234wxyz&other=3.
    The affiliate part of the URL here is affiliate_id=1234wxyz. Thus, the
    stripped URL to redirect to should be:
        https://example.com/some/page?page_id=9&other=3.
     Remember to also remove the ? if affiliate_id is the only query string
     parameter.
    :param url:
    :return:
    """
    parsed_url = urlparse.urlparse(url)
    query = urlparse.parse_qs(parsed_url.query)
    query.pop(AFFILIATE_QUERY_STRING_KEY, None)
    parsed_url = parsed_url._replace(query=urlencode(query, True))

    return urlparse.urlunparse(parsed_url)


def convert_url_query_string_to_dict(url):
    """
    Will convert the url query string into a dict, and return it.

    :param url:
    :return:
    """
    return dict(
        urlparse.parse_qsl(
            urlparse.urlsplit(url).query
        )
    )


def import_from_path(path):
    """
    Will resolve the path, and return the function or class imported.

    :param path: a Python path to a function or class
    :return: the function or class resolved
    """
    module_str, to_import_str = path.rsplit('.', 1)
    module = importlib.import_module(module_str)

    return getattr(module, to_import_str)



from django.test import TestCase


class TestAffiliateTrackingSettings(TestCase):

    def test_import_default_settings(self):
        from affiliations.affiliations_settings import (
            AFFILIATE_SESSION_KEY,
            AFFILIATE_QUERY_STRING_KEY,
        )

        self.assertEqual(
            AFFILIATE_SESSION_KEY,
            'affiliate_visitor_id'
        )
        self.assertEqual(
            AFFILIATE_QUERY_STRING_KEY,
            'affiliate_id'
        )



from django.test import TestCase
from django.contrib.auth import get_user_model
from django.utils import timezone

from affiliations.models import (
    Partner,
    Subscription,
    Visitor,
)
from affiliations.exceptions import InvalidCallBackURL

User = get_user_model()


class TestPartner(TestCase):

    def test_create(self):
        partner = Partner.objects.create(
            name='partner'
        )

        self.assertIsNotNone(partner)

    def test_default_uid(self):
        partner = Partner.objects.create(
            name='partner'
        )

        self.assertEqual(len(partner.uid), 8)

    def test_update(self):
        partner = Partner.objects.create(
            name='partner'
        )
        partner.name = 'partner updated'
        partner.save()

        self.assertEqual(partner.name, 'partner updated')

    def test_delete(self):
        partner = Partner.objects.create(
            name='partner'
        )
        partner_pk = partner.pk
        partner.delete()

        self.assertFalse(Partner.objects.filter(pk=partner_pk).exists())

    def test__str__(self):
        partner = Partner.objects.create(
            name='partner'
        )

        self.assertEqual(str(partner), partner.name)


class TestSubscription(TestCase):

    def setUp(self):
        self.partner = Partner.objects.create()

    def test_create(self):
        subscription = Subscription.objects.create(
            partner=self.partner
        )

        self.assertIsNotNone(subscription)

    def test_partner(self):
        subscription = Subscription.objects.create(
            partner=self.partner
        )

        self.assertEqual(self.partner, subscription.partner)

    def test_update(self):
        subscription = Subscription.objects.create(
            partner=self.partner
        )
        subscription.trigger = 'trigger updated'
        subscription.save()

        self.assertEqual(subscription.trigger, 'trigger updated')

    def test_delete(self):
        subscription = Subscription.objects.create(
            partner=self.partner
        )
        subscription_pk = subscription.pk
        subscription.delete()

        self.assertFalse(
            Subscription.objects.filter(pk=subscription_pk).exists()
        )

    def test__str__(self):
        subscription = Subscription.objects.create(
            partner=self.partner
        )

        self.assertEqual(str(subscription), subscription.trigger)

    def test_validate_callback_url_field_empty(self):
        subscription = Subscription.objects.create(
            partner=self.partner
        )

        with self.assertRaises(InvalidCallBackURL):
            subscription.validate_callback_url()

    def test_validate_callback_url_field_wrong_value(self):
        # The callback_url does not contain the {visitor_id} placeholder
        with self.assertRaises(InvalidCallBackURL):
            Subscription.objects.create(
                partner=self.partner,
                callback_url='http://example.com'
            )

    def test_validate_callback_url_field_right_value(self):
        right_callback_url = 'http://example.com/?{}'.format(
            Subscription.CALLBACK_URL_PLACEHOLDER
        )
        subscription = Subscription.objects.create(
            partner=self.partner,
            callback_url=right_callback_url
        )

        self.assertEqual(subscription.callback_url, right_callback_url)


class TestVisitor(TestCase):

    def setUp(self):
        self.partner = Partner.objects.create()
        self.user = User.objects.create_user(
            username='test_user',
            email='test@test.com',
            password='StrongPass123**'
        )

    def test_create(self):
        visitor = Visitor.objects.create(
            partner=self.partner,
            user=self.user
        )

        self.assertIsNotNone(visitor)

    def test_partner(self):
        visitor = Visitor.objects.create(
            partner=self.partner,
            user=self.user
        )

        self.assertEqual(visitor.partner, self.partner)

    def test_user(self):
        visitor = Visitor.objects.create(
            partner=self.partner,
            user=self.user
        )

        self.assertEqual(visitor.user, self.user)

    def test_referred_on(self):
        visitor = Visitor.objects.create(
            partner=self.partner,
            user=self.user
        )

        self.assertIsNotNone(visitor.referred_on)

    def test_entry_point(self):
        visitor = Visitor.objects.create(
            partner=self.partner,
            user=self.user,
            entry_point='http://example.com'
        )

        self.assertEqual(visitor.entry_point, 'http://example.com')

    def test_successful_on(self):
        visitor = Visitor.objects.create(
            partner=self.partner,
            user=self.user
        )

        self.assertIsNone(visitor.successful_on)

    def test_update(self):
        visitor = Visitor.objects.create(
            partner=self.partner,
            user=self.user
        )
        visitor.entry_point = 'http://example.com'
        visitor.save()

        self.assertEqual(visitor.entry_point, 'http://example.com')

    def test__str__(self):
        visitor = Visitor.objects.create(
            partner=self.partner,
            user=self.user
        )

        self.assertEqual(str(visitor), visitor.entry_point)

    def test_delete(self):
        visitor = Visitor.objects.create(
            partner=self.partner,
            user=self.user
        )
        visitor_pk = visitor.pk
        visitor.delete()

        self.assertFalse(Visitor.objects.filter(pk=visitor_pk).exists())

    def test_succeed_no_user_specified(self):
        visitor = Visitor.objects.create(
            partner=self.partner,
            user=self.user
        )

        # this must only happen if the value of successful_on is None
        self.assertIsNone(visitor.successful_on)
        visitor.succeed()
        # succeed will set the successful_on to now()
        self.assertIsNotNone(visitor.successful_on)
        # if user is specified - set the user value for the visitor. This was
        # not specified so the user should not change for the visitor
        self.assertEqual(visitor.user, self.user)

    def test_succeed_user_specified(self):
        visitor = Visitor.objects.create(
            partner=self.partner,
            user=self.user
        )

        new_user = User.objects.create_user(
            username='test_new_user',
            email='test@test.com',
            password='StrongPass123**'
        )
        # this must only happen if the value of successful_on is None
        self.assertIsNone(visitor.successful_on)
        visitor.succeed(new_user)
        # succeed will set the successful_on to now()
        self.assertIsNotNone(visitor.successful_on)
        # if user is specified - set the user value for the visitor. This was
        # not specified so the user should not change for the visitor
        self.assertEqual(visitor.user, new_user)

    def test_succeed_successful_on_not_none(self):
        now = timezone.now()
        visitor = Visitor.objects.create(
            partner=self.partner,
            user=self.user,
            successful_on=now
        )

        new_user = User.objects.create_user(
            username='test_new_user_second',
            email='test@test.com',
            password='StrongPass123**'
        )
        visitor.succeed(new_user)

        # this must only happen if the value of successful_on is None, so
        # the user, and successful_on should not change
        self.assertEqual(visitor.successful_on, now)
        self.assertEqual(visitor.user, self.user)



"""
Default settings for affiliations.

This module will include all the default settings that the affiliations
application will use.
"""
from django.conf import settings


AFFILIATE_QUERY_STRING_KEY = getattr(
    settings,
    'AFFILIATE_QUERY_STRING_KEY',
    'affiliate_id'
)

AFFILIATE_SESSION_KEY = getattr(
    settings,
    'AFFILIATE_SESSION_KEY',
    'affiliate_visitor_id'
)



"""Built in trigger receivers for the most common cases."""

from __future__ import unicode_literals

import requests
from tls import request as tls_request

from .models import Visitor
from .affiliations_settings import AFFILIATE_SESSION_KEY


def complete_trigger(request, affiliate_trigger_name):
    """
    Following are actions for when the trigger preconditions has been met.

    - Check whether the affiliate_visitor of the request has a value.
    - If the visitor is found, check that the partner of the visitor is active,
    and that it is subscribed to the trigger with the provided name. If not, do
    not proceed.
    - Do a callback to the partner subscription callback_url, after properly
    substituting the visitor_id on the URL.
    - Call succeed on the visitor, with the user of the request as an argument,
    - Unset the AFFILIATE_SESSION_KEY in the request session, as we now
    consider the affiliate reference business done.

    :param request: request argument as found via the tls module
    :param affiliate_trigger_name:
    :return:
    """
    if hasattr(request, 'affiliate_visitor'):
        visitor = Visitor.objects.filter(
            pk=request.affiliate_visitor.pk
        ).first()
        if visitor:
            partner = visitor.partner
            if partner.active:
                subscription = partner.subscriptions.filter(
                    trigger=affiliate_trigger_name
                ).first()
                if subscription:
                    # Do a callback to the partner subscription callback_url,
                    # after properly substituting the visitor_id on the URL.
                    requests.get(
                        subscription.callback_url.format(visitor_id=visitor.pk)
                    )
                    # Call succeed on the visitor, with the user of the request
                    # as an argument
                    visitor.succeed(request.user)
                    # Unset the AFFILIATE_SESSION_KEY in the request session,
                    # as we now consider the affiliate reference business done
                    del(request.session[AFFILIATE_SESSION_KEY])


def object_created(
        sender,
        instance,
        affiliate_trigger_name,
        created=None,
        **kwargs
):
    """
    Trigger for object creation.

    This trigger has a simple precondition: Check if created for the signal is
    True. If so, call the complete_trigger function, with the request argument
    as found via the tls module, along with the affiliate_trigger_name provided
    to itself.

    :param sender:
    :param instance:
    :param affiliate_trigger_name:
    :param created:
    :param kwargs:
    :return:
    """
    if created:
        complete_trigger(tls_request, affiliate_trigger_name)


def object_saved(
        sender,
        instance,
        affiliate_trigger_name,
        created=None,
        **kwargs
):
    """
    Trigger when saving an object.

    This trigger has no preconditions. Just call the complete_trigger function,
    with the request argument as found via the tls module, along with the
    affiliate_trigger_name provided to itself.

    :param sender:
    :param instance:
    :param affiliate_trigger_name:
    :param created:
    :param kwargs:
    :return:
    """
    complete_trigger(tls_request, affiliate_trigger_name)



"""The middleware regarding the affiliations application will live here."""

from __future__ import absolute_import
try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse
try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

from django.http import HttpResponseRedirect

from .models import (
    Visitor,
    Partner,
)
from .utils import (
    strip_affiliate_part_from_url,
    convert_url_query_string_to_dict,
)
from .affiliations_settings import (
    AFFILIATE_QUERY_STRING_KEY,
    AFFILIATE_SESSION_KEY,
)


class AffiliateVisitorsRegistrationMiddleware(object):
    """Middleware class to handle the affiliate visitors registrations."""

    def process_request(self, request):
        """
        Set an affiliate_visitor property on the request object.

        following these rules:
        - If AFFILIATE_SESSION_KEY is not set in the session, assign a
            value of None to the property.
        - If it is set, look up the visitor by its id. If a visitor is
            found, assign this instance as the value of the property.
        Otherwise, unset the session value, as it is obviously broken.
        """
        visitor = request.session.get(AFFILIATE_SESSION_KEY)
        if not visitor:
            request.affiliate_visitor = None
        else:
            visitor = Visitor.objects.filter(pk=visitor.pk).first()
            if visitor:
                request.affiliate_visitor = visitor
            else:
                del(request.session[AFFILIATE_SESSION_KEY])
                request.affiliate_visitor = None

    def process_response(self, request, response):
        """
        Process the response.

        following these rules:
        - Check for the existence of the AFFILIATE_QUERY_STRING_KEY in the
        query string, return properly for a middleware if not.
        - Check whether the value matches the uid of an active partner.
         If not, strip away the affiliate part of the URL and do a redirect to
         the resulting URL.
        - If an active partner is found, create a new visitor instance, set
        the entry_point to the URL of the request stripped from the
        affiliation part.
        - Set the value for AFFILIATE_SESSION_KEY in the session to the id of
        the visitor instance.
        - Set the value of the affiliate_visitor request property to the new
        visitor instance.
        - Strip away the affiliate part of the URL and do a redirect to the
        resulting URL.

        Redirects must only happen for GET requests.
        POST requests must pass for further processing.
        """
        url_requested_path = request.get_full_path()
        # - Check for the existence of the AFFILIATE_QUERY_STRING_KEY in the
        # query string, return properly for a middleware if not.
        # Redirects must only happen for GET requests.
        if (AFFILIATE_QUERY_STRING_KEY not in url_requested_path or
           request.method != 'GET'):
            return response
        else:
            active_partner = Partner.objects.filter(
                active=True,
                uid=convert_url_query_string_to_dict(
                    url_requested_path
                )[AFFILIATE_QUERY_STRING_KEY]
            ).first()
            stripped_url = strip_affiliate_part_from_url(url_requested_path)
            # An active partner found
            if active_partner:
                # Create a new visitor instance, set the entry_point to the URL
                # of the request stripped from the affiliation part.
                visitor = Visitor.objects.create(
                    partner=active_partner,
                    user=request.user,
                    entry_point=stripped_url
                )
                # Set the value for AFFILIATE_SESSION_KEY in the session to the
                # id of the visitor instance.
                request.session[AFFILIATE_SESSION_KEY] = visitor.pk
                # Set the value of the affiliate_visitor request property to
                # the new visitor instance.
                request.affiliate_visitor = visitor

            return HttpResponseRedirect(stripped_url)



from django.test import TestCase
from django.db.models import Model

from affiliations.utils import (
    strip_affiliate_part_from_url,
    convert_url_query_string_to_dict,
    import_from_path,
)


class TestUtils(TestCase):

    def test_strip_affiliate_part_from_url_just_one_param(self):
        url = 'https://example.com/some/?affiliate_id=1234wxyz'

        self.assertEqual(
            strip_affiliate_part_from_url(url),
            'https://example.com/some/'
        )

    def test_strip_affiliate_part_from_url_several_params(self):
        original_url = (
            'https://example.com/some/page?page_id=9&affiliate_id=1234wxyz'
            '&other=3'
        )
        expected_url = 'https://example.com/some/page?other=3&page_id=9'

        self.assertEqual(
            convert_url_query_string_to_dict(
                strip_affiliate_part_from_url(original_url)
            ),
            convert_url_query_string_to_dict(expected_url)
        )

    def test_convert_url_query_string_to_dict_no_params(self):
        url = 'https://example.com/some/page/'
        self.assertEqual(
            convert_url_query_string_to_dict(url),
            {}
        )

    def test_convert_url_query_string_to_dict_several_params(self):
        url = (
            'https://example.com/some/page?page_id=9&affiliate_id=1234wxyz'
            '&other=3'
        )
        self.assertEqual(
            convert_url_query_string_to_dict(url),
            {'page_id': '9', 'affiliate_id': '1234wxyz', 'other': '3'}
        )

    def test_import_from_path_module_exists(self):

        self.assertEqual(
            import_from_path('django.db.models.Model'),
            Model
        )



"""For testing the triggers module"""

from django.test import (
    TestCase,
    RequestFactory,
)
from django.contrib.auth import get_user_model

from mock import mock

from affiliations.triggers import (
    complete_trigger,
    object_created,
    object_saved,
)
from affiliations.models import (
    Visitor,
    Partner,
    Subscription,
)
from affiliations.affiliations_settings import (
    AFFILIATE_SESSION_KEY,
)


User = get_user_model()
tls_request = RequestFactory().get('/')


class TestTriggers(TestCase):

    def setUp(self):
        self.factory = RequestFactory()
        self.affiliate_trigger_name = 'Partner updated'
        self.trigger_name_object_created = 'object_created trigger name'
        self.trigger_name_object_saved = 'object_saved trigger name'

    def test_complete_trigger_affiliate_visitor_not_in_request(self):
        visitors_successful_on_none_before_complete_trigger_count = (
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )
        request = self.factory.get('/')
        complete_trigger(request, self.affiliate_trigger_name)

        # Check whether the affiliate_visitor of the request has a value.
        # If the visitor is not found do not proceed.
        self.assertNotIn('affiliate_visitor', request.__dict__.keys())
        self.assertEqual(
            visitors_successful_on_none_before_complete_trigger_count,
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )

    def test_complete_trigger_affiliate_visitor_in_request_visitor_not_found(
            self
    ):
        visitors_successful_on_none_before_complete_trigger_count = (
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )
        request = self.factory.get('/')
        request.affiliate_visitor = Visitor(pk=12345)
        complete_trigger(request, self.affiliate_trigger_name)

        # Check whether the affiliate_visitor of the request has a value.
        # If the visitor is not found do not proceed.
        self.assertIn('affiliate_visitor', request.__dict__.keys())
        self.assertEqual(
            visitors_successful_on_none_before_complete_trigger_count,
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )

    def test_complete_trigger_affiliate_visitor_in_request_partner_not_active(
            self
    ):
        partner = Partner.objects.create()
        user = User.objects.create_user(
            username='test_user',
            email='test@test.com',
            password='StrongPass123**'
        )
        visitor = Visitor.objects.create(
            partner=partner,
            user=user
        )
        visitors_successful_on_none_before_complete_trigger_count = (
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )
        request = self.factory.get('/')
        request.affiliate_visitor = visitor
        complete_trigger(request, self.affiliate_trigger_name)

        # If the visitor is found, check that the partner of the visitor is
        # active.
        # If not, do not proceed.
        self.assertIn('affiliate_visitor', request.__dict__.keys())
        self.assertEqual(
            visitors_successful_on_none_before_complete_trigger_count,
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )

    def test_complete_trigger_affiliate_visitor_in_request_not_subscribed(
            self
    ):
        partner = Partner.objects.create(active=True)
        user = User.objects.create_user(
            username='test_user',
            email='test@test.com',
            password='StrongPass123**'
        )
        visitor = Visitor.objects.create(
            partner=partner,
            user=user
        )
        visitors_successful_on_none_before_complete_trigger_count = (
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )
        request = self.factory.get('/')

        request.affiliate_visitor = visitor
        complete_trigger(request, self.affiliate_trigger_name)

        # If the visitor is found, check that the partner of the visitor is
        # active, and that it is subscribed to the trigger with the provided
        # name.
        # If not, do not proceed.
        self.assertIn('affiliate_visitor', request.__dict__.keys())
        self.assertEqual(
            visitors_successful_on_none_before_complete_trigger_count,
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )

    def test_complete_trigger_all_conditions_met(self):
        partner = Partner.objects.create(active=True)
        user = User.objects.create_user(
            username='test_user',
            email='test@test.com',
            password='StrongPass123**'
        )
        new_user = User.objects.create_user(
            username='new_test_user',
            email='test@test.com',
            password='StrongPass123**'
        )
        visitor = Visitor.objects.create(
            partner=partner,
            user=user
        )
        Subscription.objects.create(
            partner=partner,
            callback_url='http://example.com/?{}'.format(
                Subscription.CALLBACK_URL_PLACEHOLDER
            ),
            trigger=self.affiliate_trigger_name
        )
        visitors_successful_on_none_before_complete_trigger_count = (
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )
        request = self.factory.get('/')
        request.affiliate_visitor = visitor
        request.session = {
            AFFILIATE_SESSION_KEY: visitor.pk
        }
        request.user = new_user
        complete_trigger(request, self.affiliate_trigger_name)

        # If the visitor is found, check that the partner of the visitor is
        # active, and that it is subscribed to the trigger with the provided
        # name. If so:
        #
        # - Do a callback to the partner subscription callback_url, after
        # properly substituting the visitor_id on the URL.
        # - Call succeed on the visitor, with the user of the request as an
        # argument,
        # - Unset the AFFILIATE_SESSION_KEY in the request session, as we now
        # consider the affiliate reference business done.
        self.assertEqual(
            visitors_successful_on_none_before_complete_trigger_count - 1,
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )
        visitor.refresh_from_db()
        self.assertEqual(visitor.user, request.user)
        self.assertNotIn(
            AFFILIATE_SESSION_KEY,
            request.session.keys()
        )

    @mock.patch(
        'affiliations.triggers.tls_request',
        tls_request
    )
    def test_object_created(self):
        partner = Partner.objects.create(active=True)
        user = User.objects.create_user(
            username='test_user',
            email='test@test.com',
            password='StrongPass123**'
        )
        new_user = User.objects.create_user(
            username='new_test_user',
            email='test@test.com',
            password='StrongPass123**'
        )
        visitor = Visitor.objects.create(
            partner=partner,
            user=user
        )
        Subscription.objects.create(
            partner=partner,
            callback_url='http://example.com/?{}'.format(
                Subscription.CALLBACK_URL_PLACEHOLDER
            ),
            trigger=self.trigger_name_object_created
        )
        tls_request.affiliate_visitor = visitor
        tls_request.session = {
            AFFILIATE_SESSION_KEY: visitor.pk
        }
        tls_request.user = new_user

        visitors_successful_on_none_before_object_created_count = (
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )

        object_created(
            Visitor,
            Visitor.objects.first(),
            affiliate_trigger_name=self.trigger_name_object_created,
            created=True
        )

        # If the visitor is found, check that the partner of the visitor is
        # active, and that it is subscribed to the trigger with the provided
        # name. If so:
        #
        # - Do a callback to the partner subscription callback_url, after
        # properly substituting the visitor_id on the URL.
        # - Call succeed on the visitor, with the user of the request as an
        # argument,
        # - Unset the AFFILIATE_SESSION_KEY in the request session, as we now
        # consider the affiliate reference business done.
        self.assertEqual(
            visitors_successful_on_none_before_object_created_count - 1,
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )
        visitor.refresh_from_db()
        self.assertEqual(visitor.user, tls_request.user)
        self.assertNotIn(
            AFFILIATE_SESSION_KEY,
            tls_request.session.keys()
        )

    @mock.patch(
        'affiliations.triggers.tls_request',
        tls_request
    )
    def test_object_saved(self):
        partner = Partner.objects.create(active=True)
        user = User.objects.create_user(
            username='test_user',
            email='test@test.com',
            password='StrongPass123**'
        )
        new_user = User.objects.create_user(
            username='new_test_user',
            email='test@test.com',
            password='StrongPass123**'
        )
        visitor = Visitor.objects.create(
            partner=partner,
            user=user
        )
        Subscription.objects.create(
            partner=partner,
            callback_url='http://example.com/?{}'.format(
                Subscription.CALLBACK_URL_PLACEHOLDER
            ),
            trigger=self.trigger_name_object_saved
        )
        tls_request.affiliate_visitor = visitor
        tls_request.session = {
            AFFILIATE_SESSION_KEY: visitor.pk
        }
        tls_request.user = new_user

        visitors_successful_on_none_before_object_saved_count = (
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )

        object_saved(
            Visitor,
            Visitor.objects.first(),
            affiliate_trigger_name=self.trigger_name_object_saved
        )

        # If the visitor is found, check that the partner of the visitor is
        # active, and that it is subscribed to the trigger with the provided
        # name. If so:
        #
        # - Do a callback to the partner subscription callback_url, after
        # properly substituting the visitor_id on the URL.
        # - Call succeed on the visitor, with the user of the request as an
        # argument,
        # - Unset the AFFILIATE_SESSION_KEY in the request session, as we now
        # consider the affiliate reference business done.
        self.assertEqual(
            visitors_successful_on_none_before_object_saved_count - 1,
            Visitor.objects.filter(
                successful_on__isnull=True
            ).count()
        )
        visitor.refresh_from_db()
        self.assertEqual(visitor.user, tls_request.user)
        self.assertNotIn(
            AFFILIATE_SESSION_KEY,
            tls_request.session.keys()
        )



"""This module will contain the utilities we will use for the tests"""

from django.db.models.signals import post_save
from django.dispatch import receiver

from affiliations.models import Partner


@receiver(post_save)
def partner_updated(sender, instance, affiliate_trigger_name=None, **kwargs):
    """

    :param sender:
    :param instance:
    :param affiliate_trigger_name:
    :return:
    """
    if isinstance(instance, Partner):
        Partner.objects.filter(pk=instance.pk).update(active=True)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-04-13 18:10
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Partner',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uid', models.CharField(editable=False, max_length=8, unique=True)),
                ('name', models.CharField(max_length=250)),
                ('active', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('trigger', models.CharField(max_length=100)),
                ('callback_url', models.URLField()),
                ('partner', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='subscriptions', to='affiliations.Partner')),
            ],
        ),
        migrations.CreateModel(
            name='Visitor',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('referred_on', models.DateTimeField(auto_now_add=True)),
                ('entry_point', models.URLField()),
                ('successful_on', models.DateTimeField(null=True)),
                ('partner', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='visitors', to='affiliations.Partner')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='visitors', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



"""Setup for the affiliations application"""

from distutils.core import setup


setup(
    name='django-affiliate-tracking',
    packages=['affiliations'],
    include_package_data=True,
    package_data={},
    version='0.1',
    description='Application that provides a mechanism to track users and '
                'actions, to know when certain conditions are met.',
    long_description=open('README.rst', 'rt').read(),
    license='BSD License 2.0',
    author='Saxo Publish',
    author_email='publish@saxo.com',
    url='https://saxo.githost.io/publish/django-affiliate-tracking/',
    keywords=['django', 'affiliations'],
    zip_safe=False,
    install_requires=[
        'django>=1.9',
        'django-tls>=0.0.2',
        'mock>=2.0.0',
        'requests>=2.13.0',
    ],
    classifiers=[
      'Framework :: Django',
      'Intended Audience :: Developers',
      'Programming Language :: Python :: 2',
      'Programming Language :: Python :: 3'
    ],
)



from django.test import (
    TestCase,
    RequestFactory,
)
from django.contrib.auth import get_user_model
from django.http import (
    HttpResponse,
)

from affiliations.models import (
    Visitor,
    Partner,
)
from affiliations.middleware import (
    AffiliateVisitorsRegistrationMiddleware,
)
from affiliations.affiliations_settings import (
    AFFILIATE_QUERY_STRING_KEY,
    AFFILIATE_SESSION_KEY,
)

User = get_user_model()


class TestMiddleware(TestCase):

    def setUp(self):
        self.factory = RequestFactory()

    def test_affiliate_visitor_property_is_set(self):
        """
        Will check that the affiliate_visitor property is set on the request
        after the process_request is called for the middleware.
        """
        request = self.factory.get('/')
        request.session = {}
        AffiliateVisitorsRegistrationMiddleware().process_request(
            request
        )

        self.assertIn('affiliate_visitor', request.__dict__.keys())

    def test_affiliate_visitor_property_session_key_not_set(self):
        """
        Will check that the affiliate_visitor property is None after the
        process_request is called for the middleware.
        """
        request = self.factory.get('/')
        request.session = {}
        AffiliateVisitorsRegistrationMiddleware().process_request(request)

        # - If AFFILIATE_SESSION_KEY is not set in the session, assign a
        # value of None to the property.
        self.assertIsNone(request.affiliate_visitor)

    def test_affiliate_visitor_property_session_key_set_visitor_not_found(
            self
    ):
        """
        Will unset the session value because the visitor was not found, which
        means it is broken.
        """
        visitor = Visitor(pk=123)
        request = self.factory.get('/')
        request.session = {
            AFFILIATE_SESSION_KEY: visitor
        }

        # - If it is set, look up the visitor by its id. If a visitor is
        #   found, assign this instance as the value of the property.
        #   Otherwise, unset the session value, as it is obviously broken.
        self.assertIn(
            AFFILIATE_SESSION_KEY,
            request.session.keys()
        )
        AffiliateVisitorsRegistrationMiddleware().process_request(request)
        self.assertIsNone(request.affiliate_visitor)

    def test_affiliate_visitor_property_session_key_set_visitor_found(
            self
    ):
        """
        Will set the visitor to the property because the visitor was found.
        """
        partner = Partner.objects.create()
        user = User.objects.create_user(
            username='test_user',
            email='test@test.com',
            password='StrongPass123**'
        )
        visitor = Visitor.objects.create(
            partner=partner,
            user=user
        )
        request = self.factory.get('/')
        request.session = {
            AFFILIATE_SESSION_KEY: visitor
        }
        # - If it is set, look up the visitor by its id. If a visitor is
        #   found, assign this instance as the value of the property.
        self.assertIn(
            AFFILIATE_SESSION_KEY,
            request.session.keys()
        )
        AffiliateVisitorsRegistrationMiddleware().process_request(request)

        self.assertEqual(request.affiliate_visitor, visitor)

    def test_affiliate_query_not_found(self):
        """
        Check for the existence of the AFFILIATE_QUERY_STRING_KEY in the
        query string, return properly for a middleware if not.
        """
        request = self.factory.get('/')
        request.session = {}
        self.assertNotIn(
            AFFILIATE_QUERY_STRING_KEY,
            request.get_full_path()
        )
        original_response = HttpResponse()
        response = AffiliateVisitorsRegistrationMiddleware().process_response(
            request,
            original_response
        )

        self.assertEqual(response, original_response)

    def test_affiliate_query_found_not_uid_match_for_partner(self):
        """
        Check whether the AFFILIATE_QUERY_STRING_KEY matches the uid of an
        active partner. If not, strip away the affiliate part of the URL and
        do a redirect to the resulting URL.
        """
        original_response = HttpResponse()
        request = self.factory.get(
            '/?{}={}'.format(
                AFFILIATE_QUERY_STRING_KEY,
                'abcd1234'
            )
        )
        self.assertIn(
            AFFILIATE_QUERY_STRING_KEY,
            request.get_full_path()
        )
        response = AffiliateVisitorsRegistrationMiddleware().process_response(
            request,
            original_response
        )

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, '/')

    def test_affiliate_query_found_uid_match_for_partner_get_request(self):
        """
        If an active partner is found:
            - Create a new visitor instance, set the entry_point to the URL of
            the request stripped from the affiliation part.
            - Set the value for AFFILIATE_SESSION_KEY in the session to the id
            of the visitor instance.
            - Set the value of the affiliate_visitor request property to the
            new visitor instance.
            - Strip away the affiliate part of the URL and do a redirect to the
            resulting URL.

        Redirects must only happen for GET requests.
        POST requests must pass for further processing.
        """
        partner = Partner.objects.create(active=True)
        original_response = HttpResponse()
        user = User.objects.create_user(
            username='visitor_user',
            email='test@test.com',
            password='StrongPass123**'
        )
        request = self.factory.get(
            '/?{}={}'.format(
                AFFILIATE_QUERY_STRING_KEY,
                partner.uid
            )
        )
        request.user = user
        request.session = {}
        self.assertIn(
            AFFILIATE_QUERY_STRING_KEY,
            request.get_full_path()
        )
        response = AffiliateVisitorsRegistrationMiddleware().process_response(
            request,
            original_response
        )

        visitor = Visitor.objects.filter(user=user, partner=partner).first()
        # - Create a new visitor instance, set the entry_point to the URL of
        #     the request stripped from the affiliation part.
        self.assertIsNotNone(visitor)
        self.assertEqual(visitor.entry_point, '/')

        # - Set the value for AFFILIATE_SESSION_KEY in the session to the id
        # of the visitor instance.
        self.assertEqual(
            request.session[AFFILIATE_SESSION_KEY],
            visitor.pk
        )

        # - Set the value of the affiliate_visitor request property to the
        # new visitor instance.
        self.assertEqual(
            request.affiliate_visitor,
            visitor
        )

        # - Strip away the affiliate part of the URL and do a redirect to the
        # resulting URL. Redirects must only happen for GET requests.
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, '/')

    def test_affiliate_query_post_request(self):
        """
        Redirects must only happen for GET requests.
        POST requests must pass for further processing.
        """
        original_response = HttpResponse()
        request = self.factory.post(
            '/?{}={}'.format(
                AFFILIATE_QUERY_STRING_KEY,
                'abcd1234'
            )
        )
        response = AffiliateVisitorsRegistrationMiddleware().process_response(
            request,
            original_response
        )

        self.assertEqual(response, original_response)



"""Django models to handle affiliations."""

from __future__ import unicode_literals

from django.db import models
from django.utils.crypto import get_random_string
from django.conf import settings
from django.utils import timezone

from .exceptions import InvalidCallBackURL


class Partner(models.Model):
    """
    A partner is someone we make an affilate deal with.

    The partner will then generate traffic to the site.
    The initial referral will include the partner uid in the query string,
    to identify the traffic as originating from the particular partner.
    """

    uid = models.CharField(max_length=8, unique=True, editable=False)
    name = models.CharField(max_length=250)
    active = models.BooleanField(default=False)

    def __str__(self):
        """String representation for each instance of the model."""
        return self.name

    def save(self, *args, **kwargs):
        """Overriding the save method to set the uid if required."""
        if not self.uid:
            self.uid = get_random_string(8)

        super(Partner, self).save(*args, **kwargs)


class Subscription(models.Model):
    """A subscription tells which triggers a partner subscribes to."""

    CALLBACK_URL_PLACEHOLDER = '{visitor_id}'
    # the 'name' of one of the triggers defined in the settings.
    trigger = models.CharField(max_length=100)
    partner = models.ForeignKey(Partner, related_name='subscriptions')
    # the partner callback URL. Should have the placeholder
    # {visitor_id} in it somewhere, e.g. as the value for a query
    # string parameter.
    callback_url = models.URLField()

    def __str__(self):
        """String representation for each instance of the model."""
        return self.trigger

    def validate_callback_url(self):
        """
        Validate the callback_url.

        Validate the callback_url which should have the placeholder
        {visitor_id} in it somewhere, e.g. as the value for a query.
        """
        if self.CALLBACK_URL_PLACEHOLDER not in self.callback_url:
            raise InvalidCallBackURL

    def save(self, *args, **kwargs):
        """Overriding the save method to validate the callback_url."""
        if self.callback_url:
            self.validate_callback_url()

        super(Subscription, self).save(*args, **kwargs)


class Visitor(models.Model):
    """
    A visitor is soneone who gets referred to our site by a partner.

    The new middleware will detect that a request
    was caused by an affiliate partner and then register a new visitor.
    """

    partner = models.ForeignKey(Partner, related_name='visitors')
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='visitors')
    referred_on = models.DateTimeField(auto_now_add=True)
    # the URL at which the visitor entered our site
    entry_point = models.URLField()
    # tells the date the conditions of a 'success' were met, if at all
    successful_on = models.DateTimeField(null=True)

    def __str__(self):
        """String representation for each instance of the model."""
        return self.entry_point

    def succeed(self, user=None):
        """
        On success state.

        Will set the successful_on to now() and - if user is specified -
        set the user value for the visitor. But this must only happen if the
        value of successful_on is None. We don't want to change the success
        state on already successful affiliate visitors.
        """
        if not self.successful_on:
            self.successful_on = timezone.now()
            if user:
                self.user = user
            self.save()
